/*
	File:		MoreTextUtils.cp

	Contains:	Simple utilities for operating on text.

	Written by:	Pete Gontier

	Copyright:	Copyright  2000 by Apple Computer, Inc., All Rights Reserved.

	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
				("Apple") in consideration of your agreement to the following terms, and your
				use, installation, modification or redistribution of this Apple software
				constitutes acceptance of these terms.  If you do not agree with these terms,
				please do not use, install, modify or redistribute this Apple software.

				In consideration of your agreement to abide by the following terms, and subject
				to these terms, Apple grants you a personal, non-exclusive license, under Apples
				copyrights in this original Apple software (the "Apple Software"), to use,
				reproduce, modify and redistribute the Apple Software, with or without
				modifications, in source and/or binary forms; provided that if you redistribute
				the Apple Software in its entirety and without modifications, you must retain
				this notice and the following text and disclaimers in all such redistributions of
				the Apple Software.  Neither the name, trademarks, service marks or logos of
				Apple Computer, Inc. may be used to endorse or promote products derived from the
				Apple Software without specific prior written permission from Apple.  Except as
				expressly stated in this notice, no other rights or licenses, express or implied,
				are granted by Apple herein, including but not limited to any patent rights that
				may be infringed by your derivative works or by other works in which the Apple
				Software may be incorporated.

				The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
				WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
				WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
				PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
				COMBINATION WITH YOUR PRODUCTS.

				IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
				CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
				GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
				ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
				OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
				(INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
				ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	Change History (most recent first):

         <7>     11/4/00    Quinn   Added more 'STR#' routines with a new, consistent naming
                                    convention.  Updated header comment (blergh!).
         <6>     20/3/00    Quinn   Added MoreReplaceText and its support routines.  Added
                                    MoreStrLen.  Tidied up a little.
         <5>      1/3/00    Quinn   Minor Carbonation changes.
         <4>     21/4/99    Quinn   Added ValidStringListHandle.
         <3>     2/15/99    PCG     InlineGetHandleSize loses the Inline
         <2>    11/11/98    PCG     fix header
         <1>    11/10/98    PCG     first big re-org at behest of Quinn

	Old Change History (most recent first):

         <2>    10/23/98    PCG     add GetStringFromLongDouble
         <1>     6/16/98    PCG     initial checkin
*/

/////////////////////////////////////////////////////////////////

// MoreIsBetter Setup

#include "MoreSetup.h"

// Our Prototypes

#include "MoreTextUtils.h"

// Mac OS Interfaces

#include <Resources.h>
#include <PLStringFuncs.h>
#include <Fonts.h>

// MIB Prototypes

#include "MoreMemory.h"
#include "MoreQuickDraw.h"

/////////////////////////////////////////////////////////////////

pascal OSErr NewStringPtr (ConstStr255Param init, UInt8 maxSize, StringPtr *result)
{
	OSErr err = noErr;

	if (!MoreAssert (result && (!init || (*init <= maxSize))))
		err = paramErr;
	else
	{
		*result = StringPtr (NewPtr (1 + (maxSize ? maxSize : (init ? *init : 255))));

		if (!*result)
			err = MemError ( );
		else if (init && *init)
			PLstrcpy (*result,init);
		else
			**result = 0;
	}

	return err;
}

pascal OSErr GetNewStringList (short resID, tStringListP *newStringList)
{
	OSErr err = noErr;

	Handle h = GetResource ('STR#',resID);

	if (!h)
	{
		err = ResError ( );
		if (!err) err = resNotFound;
	}
	else
	{
		Size handleSize = GetHandleSize (h);
		(void) MoreAssert (MemError ( ) == noErr);

		if (handleSize < 2)
			err = paramErr;
		else
		{
			UInt16	stringCount		= ** (UInt16 **) h;
			Size	stringListSize	= sizeof (**newStringList) + (handleSize - 2) +
										(stringCount * sizeof (StringPtr));

			*newStringList = (tStringListP) NewPtr (stringListSize);

			if (!*newStringList)
				err = MemError ( );
			else
			{
				(*newStringList)->count = stringCount;

				if (stringCount)
				{
					UInt16 index = 0;

					StringPtr stringScan = (StringPtr) ((*newStringList)->list + stringCount);

					BlockMoveData (2 + *h, stringScan, handleSize - 2);

					do
					{
						(*newStringList)->list [index] = stringScan;
						stringScan += *stringScan + 1;
					}
					while (++index < stringCount);
				}
			}
		}

		ReleaseResource (h);
		(void) MoreAssert (ResError ( ) == noErr);
	}

	return err;
}

extern pascal OSStatus MoreStrListNew(Handle *strList)
	// See comment in header file.
{
	OSStatus err;

	MoreAssertQ(strList != nil);
		
	*strList = NewHandleClear(sizeof(SInt16));
	err = MoreMemError(*strList);

	return err;
}

static OSStatus CheckAndPrepStrList(Handle strList, SInt8 *s)
	// Checks whether strList is something like a string list handle
	// and marks the handle as not purgeable.  Sets *s to the
	// old handle state so that the caller can restore it
	// when done.
	//
	// Also, if the handle has been purged, calls the Resource Manager
	// LoadResource routine to attempt to unpurge it.  If that 
	// fails, spits out an error.
{
	OSStatus err;
	
	MoreAssertQ(s != nil);
	
	err = noErr;
	if (strList == nil) {
		err = paramErr;
	}
	if (err == noErr && *strList == nil) {
		
		// Ideally I would like to call HGetState and check for the
		// kHandleIsResourceBit bit before calling LoadResource.  But
		// HGetState returns an error if you call it on a purged handle,
		// so Im forced to call LoadResource unilaterally.  If the
		// handle is not a resource, Resource Manager will fail the
		// LoadResource call.
		
		LoadResource(strList);

		// LoadResource doesnt reliably set ResError (specifically, 
		// in the case where it runs out of memory), so we just recheck
		// for a purged handle.

		if (*strList == nil) {
			err = ResError();
			if (err == noErr) {
				err = memFullErr;
			}
		}
	}
	if (err == noErr) {
		*s = HGetState(strList);			MoreAssertQ( MemError() == noErr);
		HNoPurge(strList);					MoreAssertQ( MemError() == noErr);
	}
	return err;
}

static ByteCount GetOffsetOfIndexedString(Handle strList, SInt16 index)
	// Returns the offset of the index'th string in the
	// string list handle.  index must be in the range [1..count].
{
	ByteCount offset;
	ByteCount maxOffset;
	SInt16    thisIndex;
	
	MoreAssertQ(MoreStrListValidate(strList));

	if ( MoreAssert(index >= 1 && index <= **((SInt16 **) strList)) ) {
		thisIndex = 1;
		offset = sizeof(SInt16);
		maxOffset = GetHandleSize(strList);
		while (thisIndex != index && offset < maxOffset) {
			offset += ((UInt8 *) *strList)[offset] + 1;
			thisIndex += 1;
		}
		if (offset >= maxOffset) {
			offset = 0;
		}
	} else {
		offset = 0;
	}
	
	return offset;
}

static OSStatus CheckAndPrepStrListAndCalcOffset(Handle strList, SInt8 *s, SInt16 index, ByteCount *offset)
	// Does the same as CheckAndPrepStrList but also checks to see
	// whether index is within bounds (ie in [1..count]) and, if it
	// is, returns the offset of the indexth string in the handle.
{
	OSStatus err;

	MoreAssert(offset != nil);
		
	err = CheckAndPrepStrList(strList, s);
	if (err == noErr) {
		if (index < 1 || index > **((SInt16 **) strList)) {
			err = paramErr;
		}
		if (err == noErr) {
			*offset = GetOffsetOfIndexedString(strList, index);
			if (*offset == 0) {
				err = paramErr;
			}
		}
		if (err != noErr) {
			HSetState(strList, *s);
		}
	}
	return err;
}

extern pascal OSStatus MoreStrListEmpty(Handle strList)
	// See comment in header file.
{
	OSStatus err;
	SInt8    s;
	SInt16   tmpSInt16;
	
	// Dont call validate because strList doesnt have to be a valid
	// string list in order for us to empty it.
	
	err = CheckAndPrepStrList(strList, &s);
	if (err == noErr) {
		tmpSInt16 = 0;
		err = PtrToXHand(&tmpSInt16, strList, sizeof(tmpSInt16));
		
		HSetState(strList, s);
	}
	MoreAssertQ(err != noErr || MoreStrListValidate(strList));
	return err;
}

extern pascal SInt16   MoreStrListCount(Handle strList)
	// See comment in header file.
{
	OSStatus err;
	SInt8    s;
	SInt16   result;
	
	MoreAssertQ(MoreStrListValidate(strList));

	result = 0;
	err = CheckAndPrepStrList(strList, &s);
	if (err == noErr) {
		result = **((SInt16 **) strList);
		HSetState(strList, s);
	}
	MoreAssertQ(err == noErr);		// nowhere to place the error in non-debug build
	return result;
}

extern pascal OSStatus MoreStrListGetIndexed(Handle strList, SInt16 index, Str255 str)
	// See comment in header file.
{
	OSStatus  err;
	SInt8     s;
	ByteCount offset;
	
	MoreAssertQ(MoreStrListValidate(strList));
	MoreAssertQ(str != nil);
	
	err = CheckAndPrepStrListAndCalcOffset(strList, &s, index, &offset);
	if (err == noErr) {
		BlockMoveData(*strList + offset, str, ((UInt8 *) *strList)[offset] + 1);
		
		HSetState(strList, s);
	}
	return err;
}

extern pascal OSStatus MoreStrListSetIndexed(Handle strList, SInt16 index, ConstStr255Param str)
	// See comment in header file.
{
	OSStatus  err;
	SInt8     s;
	ByteCount offset;
	
	MoreAssertQ(MoreStrListValidate(strList));
	MoreAssertQ(str != nil);
	
	err = CheckAndPrepStrListAndCalcOffset(strList, &s, index, &offset);
	if (err == noErr) {
		// Munger Explained:
		// At offset bytes into the strList, search for an N byte
		// string (where N is length of the string at offset)
		// with any contents (nil) and replace with str.
		(void) Munger(strList, offset, nil, ((UInt8 *) *strList)[offset] + 1, str, str[0] + 1);
		err = MemError();
		
		HSetState(strList, s);
	}
	MoreAssertQ(MoreStrListValidate(strList));		// Even if we fail, the result should still be well-formed.
	return err;
}

extern pascal OSStatus MoreStrListDeleteIndexed(Handle strList, SInt16 index)
	// See comment in header file.
{
	OSStatus  err;
	SInt8     s;
	ByteCount offset;
	
	MoreAssertQ(MoreStrListValidate(strList));
	
	err = CheckAndPrepStrListAndCalcOffset(strList, &s, index, &offset);
	if (err == noErr) {
		// Munger Explained:
		// At offset bytes into the strList, search for an N byte
		// string (where N is length of the string at offset) with
		// any contents (nil) and replace with nothing ((Ptr) -1, 0).
		(void) Munger(strList, offset, nil, ((UInt8 *) *strList)[offset] + 1, (Ptr) -1, 0);
		err = MemError();

		if (err == noErr) {
			**((SInt16 **) strList) -= 1;
		}
		
		HSetState(strList, s);
	}
	MoreAssertQ(MoreStrListValidate(strList));		// Even if we fail, the result should still be well-formed.
	return err;
}

extern pascal OSStatus MoreStrListInsertAtIndex(Handle strList, SInt16 index, ConstStr255Param str)
	// See comment in header file.
{
	OSStatus  err;
	SInt8     s;
	ByteCount offset;
	
	MoreAssertQ(MoreStrListValidate(strList));
	MoreAssertQ(str != nil);
	
	err = CheckAndPrepStrList(strList, &s);
	if (err == noErr) {
		
		// If were adding to the end, call the append routine.
		// We have to do this check inside the CheckAndPreStrList
		// because the strList might be purged and we need to read
		// the string count from it.  Calling MoreStrListAppend (as
		// opposed to some internal routine) isnt a problem because
		// the only real effect from CheckAndPreStrList is to mark
		// the handle as non-purgeable, which is idempotent.

		if (index == ((**((SInt16 **) strList) + 1))) {
			err = MoreStrListAppend(strList, str);
		} else {
		
			// Now we know were inserting in the middle.  Do the
			// bounds check on index and then find the offset of
			// the current index'th string.
			
			if (index < 1 || index > **((SInt16 **) strList)) {
				err = paramErr;
			}
			if (err == noErr) {
				offset = GetOffsetOfIndexedString(strList, index);
				if (offset == 0) {
					err = paramErr;
				}
			}
			
			// Insert the new string in the handle and then up the string
			// count.
			
			if (err == noErr) {
				// Munger Explained:
				// At offset bytes into the strList, search for nothing
				// (nil, 0) and replace with str (str, str[0] + 1).
				(void) Munger(strList, offset, nil, 0, str, str[0] + 1);
				err = MemError();
			}
			if (err == noErr) {
				**((SInt16 **) strList) += 1;
			}
		}
		
		HSetState(strList, s);
	}
	MoreAssertQ(MoreStrListValidate(strList));		// Even if we fail, the result should still be well-formed.
	return err;
}

extern pascal OSStatus MoreStrListAppend(Handle strList, ConstStr255Param str)
	// See comment in header file.
{
	OSStatus  err;
	SInt8     s;

	MoreAssertQ(MoreStrListValidate(strList));
	MoreAssertQ(str != nil);
	
	err = CheckAndPrepStrList(strList, &s);
	if (err == noErr) {
		err = PtrAndHand(str, strList, str[0] + 1);
		if (err == noErr) {
			**((SInt16 **) strList) += 1;
		}
		
		HSetState(strList, s);
	}
	MoreAssertQ(MoreStrListValidate(strList));		// Even if we fail, the result should still be well-formed.
	return err;
}

extern pascal Boolean  MoreStrListValidate(Handle strList)
	// See comment in header file.
{
	Boolean result;
	UInt16 stringCount;
	UInt16 stringIndex;
	UInt8* cursor;
	UInt8* bound;
	ByteCount strListLength;
	SInt8  s;

	if ( CheckAndPrepStrList(strList, &s) != noErr) {
		result = false;
	} else {
		MoreAssertQ(strList != nil);			// because otherwise CheckAndPreStrList would have errored
		MoreAssertQ(*strList != nil);			// ditto
		
		result = true;
		if (GetHandleSize(strList) < sizeof(UInt16) ) {
			result = false;
		}
		if (result) {
			stringCount = *((UInt16 *) *strList);
			
			strListLength = GetHandleSize(strList);
			
			// From here down we have to make sure we do nothing to move
			// until we're done with cursor and bound.
			
			cursor = ((UInt8 *)*strList) + sizeof(UInt16);
			bound  = ((UInt8 *)*strList) + strListLength;
			for (stringIndex = 0; stringIndex < stringCount; stringIndex++) {
				if ( cursor < bound ) {
					cursor += *cursor + 1;
				} else {
					result = false;
					break;
				}
			}
			
			if (result) {
				if ( cursor != bound ) {
					result = false;
				}
			}
		}
		HSetState(strList, s);
	}
	
	return result;
}

pascal OSErr GetPascalStringFromLongDouble (long double d, SInt8 precision, StringPtr buf)
{
	OSErr err = noErr;

	*buf = 0;

	//	If client requests more precision than is available, bitch.

	if (!MoreAssert (precision < SIGDIGLEN))
		err = paramErr;
	else
	{
		// 	If precision is less than 0, client is requesting
		//	as much precision as is available.

		if (precision < 0)
			precision = SIGDIGLEN;

		//	Try normal notation, and if that overflows,
		//	fall back to scientific notation (bleah!).

		decimal dec;
		decform df = { 1, 0, precision };

		num2dec (&df,d,&dec);

		if (dec.sig.text [0] == '?')
		{
			df.style	= 0;
			df.digits	= SIGDIGLEN;

			num2dec (&df,d,&dec);
		}

		if (*(dec.sig.text) == '0')
			PLstrcpy (buf,"\p0");
		else if (*(dec.sig.text) == 'I')
			PLstrcpy (buf,"\p[INF]");
		else if (*(dec.sig.text) == 'N')
			PLstrcpy (buf,"\p[NaN]");
		else
		{
			//	Finally, convert it to a pascal string...

			dec2str (&df,&dec,(Ptr)buf);
			CopyCStringToPascal( (char *) buf, buf );

			//	...and trim trailing zeroes.

//			if (df.style == 1 && dec.exp < 0)
//				while (buf [*buf] == '0')
//					*buf -= 1;
		}
	}

	return err;
}

// gDummyPort is a pointer to a dummy GrafPort whose font is always
// the system font.  By setting thePort to this port, we can then safely
// call Script Manager that get script information from the current port.

static GrafPtr gDummyPort;

extern pascal OSStatus InitMoreTextUtils(void)
	// See comment in header.
{
	OSStatus err;
	GrafPtr  oldPort;

	// Preserve the current port.
		
	GetPort(&oldPort);
	
	// Create gDummyPort.
	
	gDummyPort = (GrafPtr) CreateNewPort();
	err = MoreMemError(gDummyPort);
	if (err == noErr) {
	
		// CreateNewPort, if it follows the semantics of the old
		// OpenPort routine, should set thePort to the new port.
		// This assert ensures that behaviour, because the following
		// call to TextFont relies on it.
		
		#if MORE_DEBUG
			{
				GrafPtr junkPort;
				
				GetPort(&junkPort);
				MoreAssertQ(junkPort == gDummyPort);
			}
		#endif

		// Set the dummy ports font to the system font.
				
		TextFont(systemFont);
	}
	
	SetPort(oldPort);
	
	return err;
}

extern pascal void TermMoreTextUtils(void)
	// See comment in header.
{
	if (gDummyPort != nil) {
		DisposePort( (CGrafPtr) gDummyPort);
		gDummyPort = nil;
	}
}

extern pascal void MoreReplaceText(Str255 baseText, ConstStr255Param substitutionText, ConstStr15Param key)
	// See comment in header.
{
	OSStatus err;
	GrafPtr  oldPort;
	Handle   baseTextH;
	Handle   substitutionTextH;
	Size     newLength;
	UInt8 	 *tmpKey;

	MoreAssertQ(baseText != nil);
	MoreAssertQ(substitutionText != nil);
	MoreAssertQ(key != nil);
	MoreAssertQ(key[0] <= 15);
	
	GetPort(&oldPort);
	
	baseTextH = nil;
	substitutionTextH = nil;

	// Copy the strings into handles, call ReplaceText, then
	// copy the result back into baseText (making sure that
	// there is at most 255 bytes in the resulting string).
		
	err = noErr;
	if (gDummyPort == nil) {
		err = InitMoreTextUtils();
	}
	if (err == noErr) {
		SetPort(gDummyPort);

		err = PtrToHand(&baseText[1], &baseTextH, baseText[0]);
	}
	if (err == noErr) {
		err = PtrToHand(&substitutionText[1], &substitutionTextH, substitutionText[0]);
	}
	if (err == noErr) {
		tmpKey = (UInt8 *) key;
		err = ReplaceText(baseTextH, substitutionTextH, tmpKey);
		if (err >= 0) {
			newLength = GetHandleSize(baseTextH);
			if (newLength > 255) {
				newLength = 255;
			}
			baseText[0] = newLength;
			BlockMoveData(*baseTextH, &baseText[1], newLength);
			err = noErr;
		}
	}
	
	// Clean up.

	if (baseTextH != nil) {
		DisposeHandle(baseTextH);
		MoreAssertQ(MemError() == noErr);
	}	
	if (substitutionTextH != nil) {
		DisposeHandle(substitutionTextH);
		MoreAssertQ(MemError() == noErr);
	}	
	SetPort(oldPort);
	
	MoreAssert(err == noErr);
}

extern pascal ByteCount MoreStrLen(const char *str)
	// See comment in header.
{
	ByteCount result;
	
	result = 0;
	while ( *str != 0 ) {
		result += 1;
		str += 1;
	}
	return result;
}
