/*
	File:		MoreMPLog.c

	Contains:	MP-safe logging.

	Written by:	Quinn

	Copyright:	Copyright  2000 by Apple Computer, Inc., All Rights Reserved.

	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
				("Apple") in consideration of your agreement to the following terms, and your
				use, installation, modification or redistribution of this Apple software
				constitutes acceptance of these terms.  If you do not agree with these terms,
				please do not use, install, modify or redistribute this Apple software.

				In consideration of your agreement to abide by the following terms, and subject
				to these terms, Apple grants you a personal, non-exclusive license, under Apples
				copyrights in this original Apple software (the "Apple Software"), to use,
				reproduce, modify and redistribute the Apple Software, with or without
				modifications, in source and/or binary forms; provided that if you redistribute
				the Apple Software in its entirety and without modifications, you must retain
				this notice and the following text and disclaimers in all such redistributions of
				the Apple Software.  Neither the name, trademarks, service marks or logos of
				Apple Computer, Inc. may be used to endorse or promote products derived from the
				Apple Software without specific prior written permission from Apple.  Except as
				expressly stated in this notice, no other rights or licenses, express or implied,
				are granted by Apple herein, including but not limited to any patent rights that
				may be infringed by your derivative works or by other works in which the Apple
				Software may be incorporated.

				The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
				WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
				WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
				PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
				COMBINATION WITH YOUR PRODUCTS.

				IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
				CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
				GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
				ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
				OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
				(INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
				ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	Change History (most recent first):

         <1>     7/11/00    Quinn   First checked in.
*/

/////////////////////////////////////////////////////////////////////

// MoreIsBetter Setup

#include "MoreSetup.h"

// Mac OS Interfaces

#include <Multiprocessing.h>
#include <DriverSynchronization.h>
#include <Gestalt.h>

// Standard C interfaces

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <FSp_fopen.h>

// MIB interfaces

#include "MoreMemory.h"

// Our prototypes

#include "MoreMPLog.h"

/////////////////////////////////////////////////////////////////////
#pragma mark ----- Core Logging Routines -----

// CFM export these data structures so that we can find them in MacsBug.

__declspec (export) extern ByteCount gMPLogBufferSize = 0;

__declspec (export) extern UInt8 *gMPLogBuffer = nil;

__declspec (export) extern UInt32 gMPLogCurrentBufferIndex = 0;

const char *gBufferTerminator = "___End Buffer___";

#if TARGET_API_MAC_CARBON

	static Boolean gRunningOnCarbonX = false;
	
	static Boolean RunningOnCarbonX(void)
	{
	    UInt32 response;
	    
	    return (Gestalt(gestaltSystemVersion, 
	                    (SInt32 *) &response) == noErr)
	                && (response >= 0x01000);
	}

#endif

extern pascal OSStatus InitMPLog(ByteCount logBufferSize)
	// See comment in header file.
{
	OSStatus err;
	
	// The log buffer size must be a power of 2.  This is because the 
	// gMPLogCurrentBufferIndex is a free running counter.  I simply 
	// mod the counter with the buffer size to get the offset into the  
	// buffer.  When gMPLogCurrentBufferIndex wraps around of the 4 GB 
	// mark, things work OK as long as the buffer size is a power of 2.
	
	MoreAssertQ((logBufferSize & (logBufferSize - 1)) == 0);
	
	// Allocate the buffer.
	
	err = noErr;
	gMPLogBufferSize = logBufferSize;
	gMPLogCurrentBufferIndex = 0;
	gMPLogBuffer = (UInt8 *) NewPtrClear(gMPLogBufferSize + strlen(gBufferTerminator));
	if (gMPLogBuffer == nil) {
		err = memFullErr;
	}
	if (err == noErr) {

		// Copy the terminator into the end of the log buffer.

		BlockMoveData(gBufferTerminator, &gMPLogBuffer[gMPLogBufferSize], strlen(gBufferTerminator));

		// If we're not building Carbon, hold the memory so that we can 
		// log at non-deferred interrupt time.  Carbon builds don't have 
		// access to HoldMemory; this isn't an issue on Mac OS X but could 
		// cause problems with CarbonLib on Mac OS 9.
		
		#if ! TARGET_API_MAC_CARBON
			err = SafeHoldMemory(gMPLogBuffer, gMPLogBufferSize);
			if (err != noErr) {
				DisposePtr( (Ptr) gMPLogBuffer);
				MoreAssertQ(MemError() == noErr);
				gMPLogBuffer = nil;
			}
		#endif
	}
	
	// If we're building a debug version, the following code will define a 
	// MacsBug macro called "MPLog" which will find the last log entry 
	// marker in the buffer (the '' character) and then display the last 
	// 256 bytes of the log.  This doesn't always work perfectly (for example, 
	// if the log has just wrapped), but it was easy to implement and it works 
	// most of the time.
	
	#if MORE_DEBUG
		if (err == noErr) {
			DebugStr("\p; mc MPLog \"f gMPLogBuffer^ gMPLogBufferSize^ '' ; dm .-100 110\" ; g");
		}
	#endif
	#if TARGET_API_MAC_CARBON
		gRunningOnCarbonX = RunningOnCarbonX();
	#endif

	return err;
}

extern pascal void TermMPLog(void)
	// See comment in header file.
{
	UInt8 *tmp;
	
	if (gMPLogBuffer != nil) {
		tmp = gMPLogBuffer;
		gMPLogBuffer = nil;
		
		#if ! TARGET_API_MAC_CARBON
			{
				OSStatus junk;
				
				junk = SafeUnholdMemory(gMPLogBuffer, gMPLogBufferSize);
				MoreAssertQ(junk == noErr);
			}
		#endif
		
		DisposePtr( (Ptr) tmp);
		MoreAssertQ(MemError() == noErr);
	}
}

extern pascal void MPLogText(const UInt8 *text, ByteCount len)
	// See comment in header file.
{
	UInt32 lastCharIndex;
	
	MoreAssertQ(gMPLogBuffer != nil);			// You forgot to call InitMPLog.
	
	// Allocate the next N bytes in the log buffer by atomically 
	// adding to the index.
	
	lastCharIndex = AddAtomic(len + 2, (SInt32 *) &gMPLogCurrentBufferIndex) + len + 2;

	// Fill in the log buffer that we allocated with the text.
	
	// Plotting this bullet is not 100% certain but it's
	// sufficiently useful in helping you find the last logged text
	// that I'm prepared to accept the uncertainty.
	
	gMPLogBuffer[lastCharIndex % gMPLogBufferSize] = '';
	lastCharIndex -= 1;
	
	gMPLogBuffer[lastCharIndex % gMPLogBufferSize] = '';
	while (len != 0) {
		len -= 1;
		lastCharIndex -= 1;
		gMPLogBuffer[lastCharIndex % gMPLogBufferSize] = text[len];
	}
	lastCharIndex -= 1;
	gMPLogBuffer[lastCharIndex % gMPLogBufferSize] = '';
}

extern pascal OSStatus MPLogWriteToFile(const FSSpec *fss)
	// See comment in header file.
{
	OSStatus err;
	FILE *output;
	SInt32 rowMax;
	SInt32 row;
	SInt32 col;
	UInt32 index;
	UInt8  ch;

	MoreAssertQ(gMPLogBuffer != nil);			// You forgot to call InitMPLog.

	err = noErr;	
	(void) FSpDelete(fss);
	output = FSp_fopen(fss, "w");
	if (output == nil) {
		err = -1;
	}
	
	// Print the log buffer to the file.  Note that the output is an 
	// exact match of the MacsBug "dm" command.  This is not a coincidence.
	
	if (err == noErr) {
		rowMax = (gMPLogBufferSize + 15) / 16;
		index = 0;
		for (row = 0; row < rowMax; row++) {
			fprintf(output, "  %08X  ", index);
			for (col = 0; col < 16; col+=2) {
				if ((index + col) < gMPLogBufferSize) {
					fprintf(output, "%04X ", *((UInt16 *) &gMPLogBuffer[index + col]));
				} else {
					fprintf(output, "     ");
				}
			}
			fprintf(output, " ");
			for (col = 0; col < 16; col++) {
				if ((index + col) < gMPLogBufferSize) {
					ch = gMPLogBuffer[index + col];
					if (ch < 32) {
						ch = '.';
					}
					fprintf(output, "%c", ch);
				} else {
					fprintf(output, " ");
				}
			}
			fprintf(output, "\n");
			index += 16;
		}
		(void) fclose(output);
	}
	return err;
}

/////////////////////////////////////////////////////////////////
#pragma mark ----- Log Points -----

#if MORE_DEBUG

	static UInt32 gLogMask = 0;

	extern pascal void MPLogSetMask(UInt32 logMask)
	{
		gLogMask = logMask;
	}

	extern pascal UInt32 MPLogGetMask(void)
	{
		return gLogMask;
	}

	// This section of code makes heavy use of sprintf.
	// I'm a little worried about this because there's no guarantee 
	// that sprintf is MP safe.  For the moment I'm going to let it 
	// slide, but it would be good to revisit.
	
	extern pascal void MPLog(UInt32  module, OSType tag)
	{
		char tmpStr[256];
		
		if ( gLogMask & (1 << module) ) {
			(void) sprintf(tmpStr, "[%08lx]%4.4s", MPCurrentTaskID(), &tag);
			MPLogText( (UInt8 *) tmpStr, strlen(tmpStr));       
		}
	}

	extern pascal void MPLog1(UInt32 module, OSType tag, const void *p1)
	{
		char tmpStr[256];
		
		if ( gLogMask & (1 << module) ) {
			(void) sprintf(tmpStr, "[%08lx]%4.4s %08lx   ", MPCurrentTaskID(), &tag, p1);
			MPLogText( (UInt8 *) tmpStr, strlen(tmpStr));       
		}
	}

	extern pascal void MPLog2(UInt32 module, OSType tag, const void *p1, const void *p2)
	{
		char tmpStr[256];
		
		if ( gLogMask & (1 << module) ) {
			(void) sprintf(tmpStr, "[%08lx]%4.4s %08lx %08lx  ", MPCurrentTaskID(), &tag, p1, p2);
			MPLogText( (UInt8 *) tmpStr, strlen(tmpStr));       
		}
	}

	extern pascal void MPLog3(UInt32 module, OSType tag, const void *p1, const void *p2, const void *p3)
	{
		char tmpStr[256];
		
		if ( gLogMask & (1 << module) ) {
			(void) sprintf(tmpStr, "[%08lx]%4.4s %08lx %08lx %08lx ", MPCurrentTaskID(), &tag, p1, p2, p3);
			MPLogText( (UInt8 *) tmpStr, strlen(tmpStr));       
		}
	}

	extern pascal void MPLog4(UInt32 module, OSType tag, const void *p1, const void *p2, const void *p3, const void *p4)
	{
		char tmpStr[256];
		
		if ( gLogMask & (1 << module) ) {
			(void) sprintf(tmpStr, "[%08lx]%4.4s %08lx %08lx %08lx %08lx", MPCurrentTaskID(), &tag, p1, p2, p3, p4);
			MPLogText( (UInt8 *) tmpStr, strlen(tmpStr));       
		}
	}

#endif

/////////////////////////////////////////////////////////////////////
#pragma mark ----- Utility printf Routines -----

extern void MPLogPrintf(const char *format, ...)
	// See comment in header file.
{
	char 	tmpStr[256];
	va_list args;                            

	va_start(args, format);

	vsprintf(tmpStr, format, args);
	MPLogText( (UInt8 *) tmpStr, strlen(tmpStr));       
	
	va_end(args);
}

// MPLogPrintfSlow uses MPRemoteCall to get back to system task 
// time in order to print to the console.  It needs multiple 
// parameters to do this, which it collects into the BluePrinterArgs 
// parameter block.

enum {
	kBluePrinterArgsMagic = 'BlPr'
};

struct BluePrinterArgs {
	OSType		 magic;			// must be kBluePrinterArgsMagic
	const char * format;
	va_list      args;
};
typedef struct BluePrinterArgs BluePrinterArgs;

static void *BluePrinter(void *parameter)
{
	BluePrinterArgs *blue;
	
	blue = (BluePrinterArgs *) parameter;
	MoreAssertQ(blue->magic == kBluePrinterArgsMagic);
	
	vprintf(blue->format, blue->args);
	fflush(stdout);
	
	return nil;
}

extern void MPLogPrintfSlow(const char *format, ...)
	// See comment in header file.
{
	BluePrinterArgs blue;
	va_list  		args;

	va_start(args, format);
	            
	blue.magic  = kBluePrinterArgsMagic;
	blue.format = format;
	blue.args   = args;	

	#if TARGET_API_MAC_CARBON
		// MPRemoteCall is broken on Carbon on Mac OS Public Beta [2544931].
		// We just call BluePrinter directly, hoping that the standard 
		// C library is sufficiently MP safe.
		if (gRunningOnCarbonX) {
			(void) BluePrinter(&blue);
		} else {
			(void) MPRemoteCall(BluePrinter, &blue, kMPOwningProcessRemoteContext);
		}
	#else
		(void) MPRemoteCall(BluePrinter, &blue, kMPOwningProcessRemoteContext);
	#endif
	
	va_end(args);
}
